;;; -*- Mode: LISP; Syntax: Common-lisp; Base: 10; Lowercase: Yes; Package: USER;  -*-

;;;; Module definition for CLX

;;; This file is a Common Lisp Module description, but you will have to edit
;;; it to meet the needs of your site.

;;; Ideally, this file (or a file that loads this file) should be
;;; located in the system directory that REQUIRE searches.  Thus a user
;;; would say
;;;			(require :clx)
;;; to load CLX.  If there is no such registry, then the user must
;;; put in a site specific
;;;			(require :clx <pathname-of-this-file>)
;;;

(in-package :common-lisp-user)

(provide :clx)
